#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import annotations

from .descriptors import *

client_t = ctypes.c_int8

INVALID_CLIENT = 0
IEEE_CLIENT = 2

COMPRESSED = 0x20
SENDCAPABILITIES = 0xF

MAX_DIMS = 8

class MsgHdr(ctypes.LittleEndianStructure):
    _pack_ = 4
    _fields_ = [
        ('msglen', ctypes.c_int),
        ('status', ctypes.c_int),
        ('length', ctypes.c_int16),
        ('nargs', ctypes.c_ubyte),
        ('descriptor_idx', ctypes.c_ubyte),
        ('message_id', ctypes.c_ubyte),
        ('dtype_id', dtype_t),
        ('client_type', client_t),
        ('ndims', ctypes.c_ubyte),
        ('dims', ctypes.c_int * MAX_DIMS)
    ]

class Message(MsgHdr):
    
    def __init__(self, dsc = None, compression_level: int = 0):

        if not isinstance(dsc, Descriptor):
            dsc = Descriptor(dsc)

        if type(dsc) is Descriptor:
            self.buffer = bytearray()
            self.length = 0
            self.dtype_id = DTYPE_MISSING

        else:
            if not isinstance(dsc, (DescriptorS, DescriptorA)):
                raise Exception('Only able to send CLASS_S and CLASS_A descriptors, use `SerializeIn`')
            
            self.buffer = dsc.pack_data()
            self.length = dsc.length
            self.dtype_id = dsc.dtype_id

            if isinstance(dsc, DescriptorA):
                self.ndims = dsc.dimct
                for i in range(self.ndims):
                    self.dims[i] = dsc.dims[i]

        self.client_type = IEEE_CLIENT

        self.msglen = ctypes.sizeof(Message) + len(self.buffer)

        if len(self.buffer) > 0 and compression_level > 0:
            import zlib

            compressed_buffer = zlib.compress(self.buffer, compression_level)
            if len(compressed_buffer) < len(self.buffer):
                original_msglen = self.msglen
                compressed_buffer = bytearray(original_msglen) + compressed_buffer

                self.client_type |= COMPRESSED
                self.buffer = compressed_buffer
                self.msglen = ctypes.sizeof(Message) + len(self.buffer)

    def pack(self):
        return bytes(self) + self.buffer

    def unpack_data(self, buffer):

        if (self.client_type & COMPRESSED) > 0:
            import zlib

            original_msglen = ctypes.c_uint32.from_buffer(buffer).value
            original_buffer_size = original_msglen - ctypes.sizeof(Message) - ctypes.sizeof(ctypes.c_uint32)
            compressed_buffer = buffer[ctypes.sizeof(ctypes.c_uint32) : ]
            
            buffer = zlib.decompress(compressed_buffer, bufsize=original_buffer_size)
            self.msglen = original_msglen

        dtype_id = self.dtype_id

        # The data for deprecated dtypes gets converted automatically
        # but for some reason, the dtype id remains the same
        if dtype_id == DTYPE_F:
            dtype_id = DTYPE_FS
        elif dtype_id == DTYPE_D:
            dtype_id = DTYPE_FT
        elif dtype_id == DTYPE_FC:
            dtype_id = DTYPE_FSC
        elif dtype_id == DTYPE_DC:
            dtype_id = DTYPE_FTC
        
        if self.ndims > 0:
            return DescriptorA.unpack_data(dtype_id, buffer, dims=self.dims[: self.ndims], length=self.length)
            
        else:
            return DescriptorS.unpack_data(dtype_id, buffer, length=self.length)



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import numpy
import getpass
import unittest

from ..connection import *
from ..functions import *

class CModTest(unittest.TestCase):

    SERVER = ''

    @classmethod
    def setUpClass(cls):
        if cls.SERVER != '':
            url = cls.SERVER

            if cls.USERNAME is not None:
                url = f'{cls.USERNAME}@{url}'
            else:
                cls.USERNAME = getpass.getuser()

            if cls.PORT is not None:
                url = f'{url}:{cls.PORT}'
            else:
                cls.PORT = 8000

            cls.conn = Connection(url)
            cls.conn.openTree('cmod', 1090909009)

    def setUp(self):
        if self.SERVER == '':
            raise unittest.SkipTest("--server and --cmod were not specified")

    def test_open_close_tree(self):

        self.conn.openTree('cmod', -1)
        self.conn.closeTree('cmod', -1)

    def test_tstart(self):

        self.assertEqual(self.conn.get('TSTART'), -4.0)

        # TSTART is DTYPE_F, this tests convert_float
        self.assertEqual(self.conn.getObject('TSTART'), -4.0)

    def test_ip(self):

        first10 = numpy.array([ 0, 0, -2.0493028, 313.1422, 2.0493028, 0, 311.0929, -2.0493028, 309.0436, 311.0929 ], dtype=numpy.float32)

        data = self.conn.get('\\IP').data()

        self.assertTrue((data[ : 10] == first10).all())

    def test_checkcamac(self):

        self.assertRaises(TdiINVCLADTY, self.conn.get, 'ADMIN.CHECKCAMAC')

        compare = Action(Dispatch(2, 'ALCDATA_ANALYSIS', 'INIT', 999, ''), EXT_FUNCTION(None, 'CheckCamacServers'), '', None, None)

        data = self.conn.getObject('ADMIN.CHECKCAMAC')

        self.assertEqual(data, compare)
    
    def test_getnci(self):

        tstart_fullpath = self.conn.get('getnci(TSTART, "FULLPATH")').data()

        self.assertEqual(tstart_fullpath, '\\CMOD::TOP:TSTART')

        all_actions = self.conn.get('getnci("***", "FULLPATH", "ACTION")').data()

        self.assertEqual(len(all_actions), 1421)

        all_actions_size = self.conn.get('size(getnci("***", "FULLPATH", "ACTION"))').data()

        self.assertEqual(len(all_actions), all_actions_size)

        all_actions_status = self.conn.get('getnci("***", "STATUS", "ACTION")').data()

        first10 = [1, 1, 1, 1, 1, 0, 0, 0, 0, 1]
        self.assertListEqual(all_actions_status[ : 10 ].tolist(), first10)



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

#
# This package provdes a "best effort" approach to adding compatability with the real MDSplus Python package
# Several classes are similar but with different names, others are missing entirely. So we map what we can,
# and can provide additional compatability mapping in the future
#

from ..connection import Connection, GetMany, PutMany

from ..descriptors import String, StringArray
from ..descriptors import Int8, Int16, Int32, Int64, Int8Array, Int16Array, Int32Array, Int64Array
from ..descriptors import List, Dictionary
from ..descriptors import Signal, Dimension, Window, Slope, Function, Conglom, Range, Action, Dispatch
from ..descriptors import Program, Routine, Procedure, Method, Dependency, Condition, WithUnits, Call, WithError
from ..functions import *

# Compatability

del Connection.tcl
del Connection.tdic
del Connection.mdstcl

@property
def hostspec(self):
    return self._url

Connection.hostspec = hostspec

from ..descriptors import Descriptor as Data
from ..descriptors import Descriptor as DescriptorNULL
from ..descriptors import Descriptor as EmptyData

from ..descriptors import UInt8 as Uint8
from ..descriptors import UInt16 as Uint16
from ..descriptors import UInt32 as Uint32
from ..descriptors import UInt64 as Uint64

from ..descriptors import UInt8Array as Uint8Array
from ..descriptors import UInt16Array as Uint16Array
from ..descriptors import UInt32Array as Uint32Array
from ..descriptors import UInt64Array as Uint64Array

from .. import exceptions as mdsExceptions

mdsExceptions.statusToException = mdsExceptions.getException

def checkStatus(status, ignore=tuple(), message=None):
    # TODO: message
    if mdsExceptions.STATUS_NOT_OK(status):
        exception = mdsExceptions.getException(status)
        if isinstance(exception, ignore):
            print(exception)
        else:
            raise exception


#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import argparse

from . import *

if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='mdsthin.test')

    parser.add_argument(
        '--server',
        default=None,
        help='Server to connect to'
    )

    parser.add_argument(
        '--port',
        default=None,
        help='Port to connect to'
    )

    parser.add_argument(
        '--username',
        default=None,
        help='Username to connect with'
    )

    parser.add_argument(
        '--ssh',
        default=False,
        action='store_const', const=True,
        help='Run SSH test cases by connecting to the --server over SSH'
    )

    parser.add_argument(
        '--cmod',
        default=False,
        action='store_const', const=True,
        help='Run C-Mod specific test'
    )

    parser.add_argument(
        '--write',
        default=False,
        action='store_const', const=True,
        help='Run Write tests by spinning up a local mdsip server'
    )

    parser.add_argument(
        'unittest',
        nargs='*',
        help='Arguments to unittest.main()'
    )

    args, extra = parser.parse_known_args()

    run_mdsthin_tests(
        server=args.server,
        port=args.port,
        username=args.username,
        ssh_tests=args.ssh,
        cmod_tests=args.cmod,
        write_tests=args.write,
        # unittest arguments
        argv=[ parser.prog, *extra],
    )




#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import getpass
import unittest

from ..connection import *
from ..functions import *

class ConnectionTest(unittest.TestCase):

    SERVER = ''
    PORT = None
    USERNAME = None
    TIMEOUT = 5
    SSH_ENABLED = False

    @classmethod
    def setUpClass(cls):
        if cls.SERVER != '':
            url = cls.SERVER

            if cls.USERNAME is not None:
                url = f'{cls.USERNAME}@{url}'
            else:
                cls.USERNAME = getpass.getuser()

            if cls.PORT is not None:
                url = f'{url}:{cls.PORT}'
            else:
                cls.PORT = 8000

            cls.conn = Connection(url)

    def setUp(self):
        if self.SERVER == '':
            raise unittest.SkipTest("--server was not specified")

    def test_tdi_types(self):

        tests = [
            { 'type': Int32,   'exp': '42',        'value': 42, },
            { 'type': UInt8,   'exp': '42BU',      'value': 42, },
            { 'type': UInt16,  'exp': '42WU',      'value': 42, },
            { 'type': UInt32,  'exp': '42LU',      'value': 42, },
            { 'type': UInt64,  'exp': '42QU',      'value': 42, },
            { 'type': Int8,    'exp': '42B',       'value': 42, },
            { 'type': Int16,   'exp': '42W',       'value': 42, },
            { 'type': Int32,   'exp': '42L',       'value': 42, },
            { 'type': Int64,   'exp': '42Q',       'value': 42, },
            { 'type': Float32, 'exp': '3.14159',   'value': 3.14159, },
            { 'type': Float32, 'exp': '3.14159F0', 'value': 3.14159, },
            { 'type': Float64, 'exp': '3.14159D0', 'value': 3.14159, },
            { 'type': Float64, 'exp': '3.14159G0', 'value': 3.14159, },
            # { 'type': Float128, 'exp': '3.14159H0', },

            { 'type': Float32, 'exp': 'F_FLOAT(3.14159)', 'value': 3.14159, },
            # { 'type': Float64, 'exp': 'D_FLOAT(3.14159)', 'value': 3.14159, },
            { 'type': Float64, 'exp': 'G_FLOAT(3.14159)', 'value': 3.14159, },

            { 'type': Int32Array, 'exp': '[ [2, 4], [6, 8] ]',     'value': numpy.array([ [2, 4], [6, 8] ], dtype=numpy.int32) },
            { 'type': Int8Array,  'exp': '[ [2B, 4B], [6B, 8B] ]', 'value': numpy.array([ [2, 4], [6, 8] ], dtype=numpy.int8) },
            { 'type': Int16Array, 'exp': '[ [2W, 4W], [6W, 8W] ]', 'value': numpy.array([ [2, 4], [6, 8] ], dtype=numpy.int16) },
            { 'type': Int32Array, 'exp': '[ [2L, 4L], [6L, 8L] ]', 'value': numpy.array([ [2, 4], [6, 8] ], dtype=numpy.int32) },
            { 'type': Int64Array, 'exp': '[ [2Q, 4Q], [6Q, 8Q] ]', 'value': numpy.array([ [2, 4], [6, 8] ], dtype=numpy.int64) },
        ]

        for info in tests:
            name = f"get('{info['exp']}')"
            with self.subTest(name):

                data = self.conn.get(info['exp'])
                self.assertEqual(type(data), info['type'])

                if type(info['value']) is float:
                    self.assertAlmostEqual(data.data(), info['value'], places=5)
                else:
                    self.assertEqual(data, info['value'])

        for info in tests:
            name = f"getObject('{info['exp']}')"
            with self.subTest(name):

                data = self.conn.getObject(info['exp'])
                self.assertEqual(type(data), info['type'])

                if type(info['value']) is float:
                    self.assertAlmostEqual(data.data(), info['value'], places=5)
                else:
                    self.assertEqual(data, info['value'])

    def test_getmany(self):

        expected_result = Descriptor({
            String('a'): Descriptor({ String('value'): Int32(42) }),
            String('b'): Descriptor({ String('value'): String('Hello, World!') }),
            String('c'): Descriptor({ String('error'): String('%TREE-W-NOT_OPEN, Tree not currently open') }),
        })

        gm = self.conn.getMany()
        gm.append('a', '42')
        gm.append('b', '"Hello, World!"')
        gm.append('c', 'asdf')
        result = gm.execute()

        self.assertEqual(result, expected_result)
        self.assertEqual(gm.get('a'), 42)
        self.assertEqual(gm.get('b'), 'Hello, World!')
        self.assertRaises(TreeNOT_OPEN, gm.get, 'c')

    def test_root_whoami(self):

        root_conn = Connection(f'root@{self.SERVER}')

        whoami = root_conn.get('whoami()').data()
        self.assertEqual(whoami, 'nobody', msg='Claiming to be root should map you to nobody.')

    # We're a little limited in what we can test here
    def test_ssh_subprocess(self):
        if not self.SSH_ENABLED:
            raise unittest.SkipTest('--ssh was not specified')

        # We can only test a connection string without a username if the username is default
        if self.USERNAME == getpass.getuser():
            url = f'ssh://{self.SERVER}'
            with Connection(url, timeout=self.TIMEOUT) as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}") failed'
                )

        url = f'ssh://{self.USERNAME}@{self.SERVER}'
        with Connection(url, timeout=self.TIMEOUT, ssh_port=22) as test_conn:
            whoami = test_conn.get('whoami()').data()
            self.assertEqual(
                whoami, self.USERNAME,
                f'Connection("{url}", ssh_port=22) failed'
            )

        url = f'ssh://{self.USERNAME}@{self.SERVER}:22'
        with Connection(url, timeout=self.TIMEOUT) as test_conn:
            whoami = test_conn.get('whoami()').data()
            self.assertEqual(
                whoami, self.USERNAME,
                f'Connection("{url}") failed'
            )

        # TODO: Test ssh_subprocess_args
        # TODO: Test plink

    def test_ssh_paramiko(self):
        if not self.SSH_ENABLED:
            raise unittest.SkipTest('--ssh was not specified')
        
        try:
            import warnings

            # Silence the cryptography deprecation warnings from simply importing paramiko
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                import paramiko

            # We can only test a connection string without a username if the username is default
            if self.USERNAME == getpass.getuser():
                url = f'ssh://{self.SERVER}'
                with Connection(url, timeout=self.TIMEOUT, ssh_backend='paramiko') as test_conn:
                    whoami = test_conn.get('whoami()').data()
                    self.assertEqual(
                        whoami, self.USERNAME,
                        f'Connection("{url}", ssh_backend="paramiko") failed'
                    )

            url = f'ssh://{self.USERNAME}@{self.SERVER}'
            with Connection(url, timeout=self.TIMEOUT, ssh_port=22, ssh_backend='paramiko') as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}", ssh_port=22, ssh_backend="paramiko") failed'
                )

            url = f'ssh://{self.USERNAME}@{self.SERVER}:22'
            with Connection(url, timeout=self.TIMEOUT, ssh_backend='paramiko') as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}", ssh_backend="paramiko") failed'
                )

            # TODO: Test ssh_paramiko_options

        except ImportError:
            self.skipTest('Cannot import paramiko')

    def test_sshp_subprocess(self):
        if not self.SSH_ENABLED:
            raise unittest.SkipTest('--ssh was not specified')

        # We can only test a connection string without a username if the username is default
        if self.USERNAME == getpass.getuser():
            url = f'sshp://{self.SERVER}'
            with Connection(url, timeout=self.TIMEOUT) as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}") failed'
                )

        # We can only test a connection string without a port if the port is 8000
        if self.PORT == 8000:
            url = f'sshp://{self.USERNAME}@{self.SERVER}'
            with Connection(url, timeout=self.TIMEOUT, ssh_port=22) as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}", ssh_port=22) failed'
                )

        url = f'sshp://{self.USERNAME}@{self.SERVER}:{self.PORT}'
        with Connection(url, timeout=self.TIMEOUT) as test_conn:
            whoami = test_conn.get('whoami()').data()
            self.assertEqual(
                whoami, self.USERNAME,
                f'Connection("{url}") failed'
            )

        url = f'sshp://{self.USERNAME}@{self.SERVER}:{self.PORT}'
        with Connection(url, timeout=self.TIMEOUT, ssh_port=22) as test_conn:
            whoami = test_conn.get('whoami()').data()
            self.assertEqual(
                whoami, self.USERNAME,
                f'Connection("{url}", ssh_port=22) failed'
            )
            
        # TODO: Test plink

    def test_sshp_paramiko(self):
        if not self.SSH_ENABLED:
            raise unittest.SkipTest('--ssh was not specified')
        
        try:
            import warnings

            # Silence the cryptography deprecation warnings from simply importing paramiko
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                import paramiko

            # We can only test a connection string without a username if the username is default
            if self.USERNAME == getpass.getuser():
                url = f'sshp://{self.SERVER}'
                with Connection(url, timeout=self.TIMEOUT, ssh_backend='paramiko') as test_conn:
                    whoami = test_conn.get('whoami()').data()
                    self.assertEqual(
                        whoami, self.USERNAME,
                        f'Connection("{url}", ssh_backend="paramiko") failed'
                    )

            # We can only test a connection string without a port if the port is 8000
            if self.PORT == 8000:
                url = f'sshp://{self.USERNAME}@{self.SERVER}'
                with Connection(url, timeout=self.TIMEOUT, ssh_port=22, ssh_backend='paramiko') as test_conn:
                    whoami = test_conn.get('whoami()').data()
                    self.assertEqual(
                        whoami, self.USERNAME,
                        f'Connection("{url}", ssh_port=22, ssh_backend="paramiko") failed'
                    )

            url = f'sshp://{self.USERNAME}@{self.SERVER}:{self.PORT}'
            with Connection(url, timeout=self.TIMEOUT, ssh_backend='paramiko') as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}", ssh_backend="paramiko") failed'
                )

            url = f'sshp://{self.USERNAME}@{self.SERVER}:{self.PORT}'
            with Connection(url, timeout=self.TIMEOUT, ssh_port=22, ssh_backend='paramiko') as test_conn:
                whoami = test_conn.get('whoami()').data()
                self.assertEqual(
                    whoami, self.USERNAME,
                    f'Connection("{url}", ssh_port=22, ssh_backend="paramiko") failed'
                )

        except ImportError:
            self.skipTest('Cannot import paramiko')

    def test_empty_get(self):

        result = self.conn.get('')
        self.assertEqual(type(result), Descriptor)
        self.assertEqual(result.data(), None)




#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import unittest

from ..descriptors import *
from ..functions import *

class SerializeTest(unittest.TestCase):

    def test_pack_unpack(self):

        INTEGER_ARRAY = [ 1, 1, 2, 4, 6, 10 ]
        FLOAT_ARRAY = [ -2.0, -1.0, 0.0, 1.0, 2.0, 3.0 ]

        INTEGER_MATRIX = [ [2, 4], [6, 8], [16, 32] ]
        FLOAT_MATRIX = [ [1.0, 1.5], [2.0, 2.5], [3.0, 3.5] ]

        tests = [
            # size = sizeof(mdsdsc_s_t) + sizeof(TYPE)
            { 'data': UInt8(42),        'size': 8 + 1 },
            { 'data': UInt16(42),       'size': 8 + 2 },
            { 'data': UInt32(42),       'size': 8 + 4 },
            { 'data': UInt64(42),       'size': 8 + 8 },
            { 'data': Int8(42),         'size': 8 + 1 },
            { 'data': Int16(42),        'size': 8 + 2 },
            { 'data': Int32(42),        'size': 8 + 4 },
            { 'data': Int64(42),        'size': 8 + 8 },
            { 'data': Float32(3.14159), 'size': 8 + 4 },
            { 'data': Float64(3.14159), 'size': 8 + 8 },

            # size = sizeof(mdsdsc_a_t) + (sizeof(TYPE) * len(DATA.flatten()))
            { 'data': UInt8Array(INTEGER_ARRAY),  'size': 16 + 6 },
            { 'data': UInt16Array(INTEGER_ARRAY), 'size': 16 + (6 * 2) },
            { 'data': UInt32Array(INTEGER_ARRAY), 'size': 16 + (6 * 4) },
            { 'data': UInt64Array(INTEGER_ARRAY), 'size': 16 + (6 * 8) },
            { 'data': Int8Array(INTEGER_ARRAY),   'size': 16 + 6 },
            { 'data': Int16Array(INTEGER_ARRAY),  'size': 16 + (6 * 2) },
            { 'data': Int32Array(INTEGER_ARRAY),  'size': 16 + (6 * 4) },
            { 'data': Int64Array(INTEGER_ARRAY),  'size': 16 + (6 * 8) },
            { 'data': Float32Array(FLOAT_ARRAY),  'size': 16 + (6 * 4) },
            { 'data': Float64Array(FLOAT_ARRAY),  'size': 16 + (6 * 8) },

            # size = sizeof(mdsdsc_a_t) + (sizeof(uint32) * (len(DATA.shape) + 1)) + (sizeof(TYPE) * len(DATA.flatten()))
            { 'data': UInt8Array(INTEGER_MATRIX),  'size': 16 + 12 + 6 },
            { 'data': UInt16Array(INTEGER_MATRIX), 'size': 16 + 12 + (6 * 2) },
            { 'data': UInt32Array(INTEGER_MATRIX), 'size': 16 + 12 + (6 * 4) },
            { 'data': UInt64Array(INTEGER_MATRIX), 'size': 16 + 12 + (6 * 8) },
            { 'data': Int8Array(INTEGER_MATRIX),   'size': 16 + 12 + 6 },
            { 'data': Int16Array(INTEGER_MATRIX),  'size': 16 + 12 + (6 * 2) },
            { 'data': Int32Array(INTEGER_MATRIX),  'size': 16 + 12 + (6 * 4) },
            { 'data': Int64Array(INTEGER_MATRIX),  'size': 16 + 12 + (6 * 8) },
            { 'data': Float32Array(FLOAT_MATRIX),  'size': 16 + 12 + (6 * 4) },
            { 'data': Float64Array(FLOAT_MATRIX),  'size': 16 + 12 + (6 * 8) },

            # size = sizeof(mdsdsc_s_t) + len(data)
            { 'data': String('Hello, World!'), 'size': 8 + 13 },

            # size = sizeof(mdsdsc_a_t) + pad_data_to_same_length(data)
            { 'data': StringArray([ 'one', 'three', 'seventeen' ]), 'size': 16 + 27 },
            { 'data': StringArray([ ['a', 'b'], ['one', 'three'], ['', 'seventeen'] ]), 'size': 16 + 12 + 54 },
        ]

        for info in tests:
            name = repr(info['data'])
            with self.subTest(name):
                
                buffer = info['data'].pack()
                self.assertEqual(len(buffer), info['size'])

                data_out = Descriptor.unpack(buffer)
                self.assertEqual(data_out, info['data'])


    def test_buffers(self):

        # Packed with MDSplus.Data(...).serialize().data().tobytes()
        tests = [
            { 'data': String('Hello, World!'), 'buffer': b'\r\x00\x0e\x01\x08\x00\x00\x00Hello, World!' },
            { 'data': StringArray(['one', 'seven', 'nineteen']), 'buffer': b'\x08\x00\x0e\x04\x10\x00\x00\x00\x00\x000\x01\x18\x00\x00\x00one     seven   nineteen' },

            { 'data': UInt8(42),  'buffer': b'\x01\x00\x02\x01\x08\x00\x00\x00*' },
            { 'data': UInt16(42), 'buffer': b'\x02\x00\x03\x01\x08\x00\x00\x00*\x00' },
            { 'data': UInt32(42), 'buffer': b'\x04\x00\x04\x01\x08\x00\x00\x00*\x00\x00\x00' },
            { 'data': UInt64(42), 'buffer': b'\x08\x00\x05\x01\x08\x00\x00\x00*\x00\x00\x00\x00\x00\x00\x00' },

            { 'data': Int8(42),  'buffer': b'\x01\x00\x06\x01\x08\x00\x00\x00*' },
            { 'data': Int16(42), 'buffer': b'\x02\x00\x07\x01\x08\x00\x00\x00*\x00' },
            { 'data': Int32(42), 'buffer': b'\x04\x00\x08\x01\x08\x00\x00\x00*\x00\x00\x00' },
            { 'data': Int64(42), 'buffer': b'\x08\x00\t\x01\x08\x00\x00\x00*\x00\x00\x00\x00\x00\x00\x00' },

            { 'data': Float32(3.14159), 'buffer': b'\x04\x004\x01\x08\x00\x00\x00\xd0\x0fI@' },
            { 'data': Float64(3.14159), 'buffer': b'\x08\x005\x01\x08\x00\x00\x00n\x86\x1b\xf0\xf9!\t@' },

            { 'data': EXT_FUNCTION(None, 'test_func'), 'buffer': b'\x02\x00\xc7\xc2\x14\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x16\x00\x00\x00\xa2\x00\t\x00\x0e\x01\x08\x00\x00\x00test_func' },
            {
                'data': Signal(MULTIPLY(Float32(1000.0), dVALUE()), UInt16Array([1, 2, 3, 4, 5]), UInt64Array([0, 10, 20, 30, 40])),
                'buffer': b'\x00\x00\xc3\xc2\x00\x00\x00\x00\x03\x00\x00\x00\x18\x00\x00\x00H\x00\x00\x00b\x00\x00\x00\x02\x00\xc7\xc2\x14\x00\x00\x00\x02\x00\x00\x00\x16\x00\x00\x00"\x00\x00\x00\xf7\x00\x04\x004\x01\x08\x00\x00\x00\x00\x00zD\x02\x00\xc7\xc2\x0c\x00\x00\x00\x00\x00\x00\x00\x1e\x00\x02\x00\x03\x04\x10\x00\x00\x00\x00\x000\x01\n\x00\x00\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x08\x00\x05\x04\x10\x00\x00\x00\x00\x000\x01(\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\n\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00\x00\x00\x00\x00\x1e\x00\x00\x00\x00\x00\x00\x00(\x00\x00\x00\x00\x00\x00\x00',
            },
        ]

        for info in tests:
            name = repr(info['data'])
            with self.subTest(name):
                
                data = Descriptor.unpack(info['buffer'])
                self.assertEqual(type(data), type(info['data']))
                self.assertEqual(data, info['data'])

                buffer = data.pack()
                self.assertEqual(buffer, info['buffer'])



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import argparse

from ..connection import *

if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='mdsthin.tdic')

    parser.add_argument(
        'url',
        help='URL to connect to. Allowed protocols are `tcp://`, `tcp6://`, `ssh://` and `sshp://`, defaults to `tcp://`.'
    )

    parser.add_argument(
        '--timeout',
        default=None,
        required=False,
        help='The timeout for all socket operations in seconds, defaults to 60s.'
    )

    parser.add_argument(
        '--verbose', '-v',
        action='store_const',
        const=True,
        default=False,
        required=False,
        help='Enable debug logging.'
    )

    parser.add_argument(
        '--ssh-backend',
        default=None,
        required=False,
        help='The backend implementation of ssh to use. Can be either "subprocess" or "paramiko", defaults to "subprocess".'
    )

    parser.add_argument(
        '--ssh-port',
        default=None,
        required=False,
        help='The port to ssh to when using one of the SSH protocols.'
    )

    parser.add_argument(
        '--sshp-host',
        default=None,
        required=False,
        help='The host to netcat to when using `sshp://`, defaults to "localhost".'
    )

    args = parser.parse_args()

    kwargs = {}

    if args.timeout is not None:
        kwargs['timeout'] = args.timeout

    if args.verbose:
        kwargs['verbose'] = True

    if args.ssh_backend is not None:
        kwargs['ssh_backend'] = args.ssh_backend

    if args.ssh_port is not None:
        kwargs['ssh_port'] = args.ssh_port

    if args.sshp_host is not None:
        kwargs['sshp_host'] = args.sshp_host

    print('Connecting to:', args.url)
    c = Connection(args.url, **kwargs)
    c.tdic()
    c.disconnect()



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import ctypes

###
### class_t
###

class_t = ctypes.c_ubyte

CLASS_MISSING = 0
"""Invalid class"""

CLASS_S = 1
"""`S`tatic fixed-length descriptor"""

CLASS_D = 2
"""`D`ynamic string descriptor"""

CLASS_V = 3
"""(deprecated) VMS `V`ariable buffer descriptor"""

CLASS_A = 4
"""`A`rray descriptor"""

CLASS_P = 5
"""(deprecated?) Procedure descriptor"""

CLASS_PI = 6
"""(deprecated?) Procedure incarnation descriptor"""

CLASS_J = 7
"""(deprecated) VMS debugger label descriptor"""

CLASS_JI = 8
"""(deprecated) VMS debugger label incarnation descriptor"""

CLASS_SD = 9
"""(deprecated?) Decimal string descriptor"""

CLASS_NCA = 10
"""(deprecated?) `N`on`c`ontiguous `A`rray descriptor"""

CLASS_VS = 11
"""(deprecated?) `V`arying `S`tring descriptor"""

CLASS_VSA = 12
"""(deprecated?) `V`arying `S`tring `A`rray descriptor"""

CLASS_UBS = 13
"""(deprecated?) `U`naligned `B`it `S`tring descriptor"""

CLASS_UBA = 14
"""(deprecated?) `U`naligned `B`it `A`rray descriptor"""

CLASS_SB = 15
"""(deprecated?) `S`tring with `B`ounds descriptor"""

CLASS_UBSB = 16
"""(deprecated?) `U`naligned `B`it `S`tring with `B`ounds descriptor"""

CLASS_XD = 192
"""E`x`tended `D`ynamic descriptor"""

CLASS_XS = 193
"""E`x`tended `S`tatic descriptor"""

CLASS_R = 194
"""`R`ecord descriptor"""

CLASS_CA = 195
"""`C`ompressed `A`rray descriptor"""

CLASS_APD = 196
"""`A`rray of `P`ointers to `D`ata descriptor"""

ARRAY_CLASSES = [
    CLASS_A,
    CLASS_CA,
    CLASS_APD,
]

def class_is_array(class_id) -> bool:
    if type(class_id) is class_t:
        class_id = class_id.value

    return class_id in ARRAY_CLASSES

def class_to_string(class_id) -> str:
    VALUES = {
        CLASS_MISSING: 'CLASS_MISSING',
        CLASS_S: 'CLASS_S',
        CLASS_D: 'CLASS_D',
        CLASS_V: 'CLASS_V',
        CLASS_A: 'CLASS_A',
        CLASS_P: 'CLASS_P',
        CLASS_PI: 'CLASS_PI',
        CLASS_J: 'CLASS_J',
        CLASS_JI: 'CLASS_JI',
        CLASS_SD: 'CLASS_SD',
        CLASS_NCA: 'CLASS_NCA',
        CLASS_VS: 'CLASS_VS',
        CLASS_VSA: 'CLASS_VSA',
        CLASS_UBS: 'CLASS_UBS',
        CLASS_UBA: 'CLASS_UBA',
        CLASS_SB: 'CLASS_SB',
        CLASS_UBSB: 'CLASS_UBSB',
        CLASS_XD: 'CLASS_XD',
        CLASS_XS: 'CLASS_XS',
        CLASS_R: 'CLASS_R',
        CLASS_CA: 'CLASS_CA',
        CLASS_APD: 'CLASS_APD',
    }

    if type(class_id) is class_t:
        class_id = class_id.value

    if class_id in VALUES:
        return VALUES[class_id]

    return f'CLASS_UNKNOWN({class_id})'



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import sys
import time
import ctypes
import socket
import getpass
import logging

from .message import *
from .exceptions import *
from .functions import *

INVALID_MESSAGE_ID = 0

MDSIP_VERSION = 3

SSH_BACKEND_SUBPROCESS = 'subprocess'
SSH_BACKEND_PARAMIKO   = 'paramiko'

class Connection:
    """Implements an MDSip connection to an MDSplus server."""

    def __init__(self,
        url: str,
        timeout: float = 60.0,
        verbose: bool = False,
        ssh_backend: str = SSH_BACKEND_SUBPROCESS,
        ssh_port: int = None,
        sshp_host: str = 'localhost',
        ssh_subprocess_args: list = None,
        ssh_paramiko_options: dict = None,
        ssh_use_plink: bool = False,
    ):
        """
        Initialize an MDSplus connection to a given URL.

        The URL will be parsed as `proto://username@host:port`. The default `proto` is tcp,
        the default `username` is `os.getlogin()`, and the default `port` is 8000.

        For SSH connections you can also specify `ssh_port` as a kwarg, the default is up to the
        backend, but should always be 22. Except for `ssh://`, where `ssh_port` will default to
        `port` if it is specified, otherwise it will be up to the backend. For `sshp://` you can
        also specify `sshp_host` to specify the MDSip host to netcat to after connecting to ssh,
        otherwise it will default to `localhost`.

        Supported protocols:
         * tcp:// - Connect directly to the MDSip server at `host:port` over IPv4.
         * tcp6:// - Connect directly to the MDSip server at `host:port` over IPv6.
         * ssh:// - Connect over SSH to `host:ssh_port` and then spawn `mdsip-server-ssh`.
         * sshp:// - Connect over SSH to `host:ssh_port` and then spawn `nc $sshp_host $port`.

        :param str url: The URL to connect to.
        :param float timeout: The timeout for all socket operations in seconds, defaults to 60s
        :param bool verbose: Enable debug logging for this connection.
        :param str ssh_backend: The backend implementation of ssh to use. Can be either
            'subprocess' or 'paramiko', defaults to 'subprocess'.
        :param int ssh_port: The port to ssh to when using one of the SSH protocols.
        :param str sshp_host: The host to netcat to when using `sshp://`, defaults to 'localhost'.
        :param list ssh_subprocess_args: Additional arguments to pass to the ssh subprocess
            command line when using `SSH_BACKEND_SUBPROCESS` and one of the SSH protocols.
        :param dict ssh_paramiko_options: Additional kwargs to pass to the paramiko connect()
            function when using `SSH_BACKEND_PARAMIKO` and one of the SSH protocols.
        :param bool ssh_use_plink: Attempt to use `plink.exe -batch` instead of `ssh.exe`
            for ssh:// and sshp:// connections. Remember to use `ssh_subprocess_args` to pass any
            necessary arguments.
        :raises TimeoutError: if the connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko socket wrapper fails.
        :raises paramiko.ssh_exception.*: if the paramiko client fails.
        :raises socket.gaierror: if `host` could not be resolved to an IP.
        :raises MdsException: if an unsupported protocol is specified, or if the login fails.
        """

        logging.basicConfig()

        self._logger = logging.getLogger(__name__)
        if verbose:
            self._logger.setLevel(logging.DEBUG)
        else:
            self._logger.setLevel(logging.WARNING)

        self._socket = None
        self._timeout = timeout
        self._message_id = INVALID_MESSAGE_ID
        self._server_version = None
        self._compression_level = None

        self._ssh_backend = ssh_backend
        self._ssh_port = ssh_port
        self._sshp_host = sshp_host
        self._ssh_subprocess_args = ssh_subprocess_args
        self._ssh_paramiko_options = ssh_paramiko_options
        self._ssh_use_plink = ssh_use_plink

        self._url = url
        self._host = url

        if '://' in self._host:
            self._protocol, self._host = self._host.split('://', maxsplit=1)
        else:
            self._protocol = 'tcp'

        SUPPORTED_PROTOCOLS = ['tcp', 'tcp6', 'ssh', 'sshp']
        if self._protocol not in SUPPORTED_PROTOCOLS:
            raise MdsException(f'Only the following protocols are supported: {", ".join(SUPPORTED_PROTOCOLS)}')

        if '@' in self._host:
            # We use rsplit to allow usernames connection strings like:
            # 'user@example.com@server:port'
            self._username, self._host = self._host.rsplit('@', maxsplit=1)
        else:
            # The username used for the MDSip login packet, and SSH if the protocol is `ssh://` or `sshp://`
            self._username = getpass.getuser()

        if ':' in self._host:
            self._host, self._port = self._host.split(':', maxsplit=1)
            self._port = int(self._port)

            if self._protocol == 'ssh' and self._ssh_port is None:
                self._ssh_port = self._port
        else:
            # The MDSip default port
            self._port = 8000
        self.connect()

    def __del__(self):
        self.disconnect()

    # Used for with statement
    def __enter__(self):
        return self

    # Used for with statement
    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.disconnect()

    def connect(self):
        """
        Initialize the socket and do the login handshake.

        :raises TimeoutError: if the connection fails.
        :raises socket.gaierror: if the hostname could not be resolved.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko socket wrapper fails.
        :raises paramiko.ssh_exception.*: if the paramiko client fails.
        :raises MdsException: if the login fails.
        """

        if self._protocol in ['tcp', 'tcp6']:

            if self._protocol == 'tcp':
                socket_family = socket.AF_INET
                socket_type = socket.SOCK_STREAM
            else:
                socket_family = socket.AF_INET6
                socket_type = socket.SOCK_STREAM

            self._socket = socket.socket(socket_family, socket_type)

            self._logger.debug(f'Setting timeout to {self._timeout}s')
            self._socket.settimeout(self._timeout)

            # This causes a massive speed increase
            # It was designed to reduce the number of small packets on the wire, but we use
            # small packets for a lot of things, so it only hurts us
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            # Regular MDSplus also sets SO_KEEPALIVE and SO_OOBINLINE

            self._logger.debug(f'Resolving {self._host}')
            address_list = socket.getaddrinfo(self._host, self._port, family=socket_family, type=socket_type)
            self._address = address_list[0][4]

            self._logger.debug(f'Connecting to {self._address}:{self._port}')
            self._socket.connect(self._address)

        elif self._protocol in ['ssh', 'sshp']:

            if self._protocol == 'ssh':
                command = '/bin/sh -l -c mdsip-server-ssh'
            elif self._protocol == 'sshp':
                command = f'nc {self._sshp_host} {self._port}'

            if self._ssh_backend == SSH_BACKEND_SUBPROCESS:

                import shutil
                import subprocess

                if self._ssh_use_plink:
                    ssh = shutil.which('plink')
                    if ssh is None:
                        raise Exception('Unable to find plink.exe')
                else:
                    ssh = shutil.which('ssh')
                    if ssh is None:
                        raise Exception('Unable to find ssh command')

                ssh_command = [ssh]

                if self._ssh_port is not None:
                    if self._ssh_use_plink:
                        ssh_command.append(f'-P{self._ssh_port}')
                    else:
                        ssh_command.append(f'-p{self._ssh_port}')

                if self._ssh_use_plink:
                    ssh_command.append('-batch')

                if self._ssh_subprocess_args is not None:
                    ssh_command.extend(self._ssh_subprocess_args)
                
                ssh_command.append(f'{self._username}@{self._host}')
                ssh_command.append(command)

                ssh_command_print = [ f'"{v}"' if ' ' in v else v for v in ssh_command ]
                self._logger.debug(f'Executing {" ".join(ssh_command_print)}')

                self._ssh_subprocess = subprocess.Popen(
                    ssh_command,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    # Do not pipe stderr>stdout or the first packet we recv could be the shell errors from ssh
                    stderr=subprocess.PIPE,
                )

                import threading

                # Why is adding a timeout so hard
                class SubprocessTimeout:
                    def __init__(self, proc, timeout):
                        self._proc = proc
                        self._timer = threading.Timer(interval=timeout, function=self.do_timeout)

                    def __enter__(self):
                        self._timer.start()
                        return self

                    def __exit__(self, exc_type, exc_value, exc_traceback):
                        self._timer.cancel()
                        del self._timer
                    
                    def do_timeout(self):
                        self._proc.terminate()

                class SubprocessSocket:
                    def __init__(self, proc: subprocess.Popen, timeout: float):
                        self._proc = proc
                        self._timeout = timeout

                    def recv(self, size, flags):
                        with SubprocessTimeout(self._proc, self._timeout):
                            return self._proc.stdout.read(size)
                        
                    def recv_into(self, buffer, size, flags):
                        buffer[ : size] = self.recv(size, flags)
                        return size

                    def sendall(self, buffer):
                        with SubprocessTimeout(self._proc, self._timeout):
                            self._proc.stdin.write(buffer)
                            self._proc.stdin.flush()

                    def close(self):
                        self._proc.terminate()
                        self._proc.stdin.close()
                        self._proc.stdout.close()
                        self._proc.stderr.close()
                        self._proc.wait()

                # If stderr is blocking, then we can't check it without locking the whole program up
                try:
                    os.set_blocking(self._ssh_subprocess.stderr.fileno(), False)

                    time.sleep(0.5) # Give the subprocess time to connect

                    for line in self._ssh_subprocess.stderr.readlines():
                        self._logger.warning(line.decode().rstrip())

                except:
                    self._logger.debug('Unable to change SSH subprocess stderr to non-blocking')

                # TODO: Support timeouts

                self._socket = SubprocessSocket(self._ssh_subprocess, self._timeout)

            elif self._ssh_backend == SSH_BACKEND_PARAMIKO:

                import warnings

                # Silence the cryptography deprecation warnings from simply importing paramiko
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    import paramiko

                paramiko_options = {
                    'username': self._username,
                }

                if self._ssh_port is not None:
                    paramiko_options['port'] = int(self._ssh_port)

                if self._ssh_paramiko_options is not None:
                    paramiko_options.update(self._ssh_paramiko_options)

                paramiko_options_print = [ f"{k}={repr(v)}" for k,v in paramiko_options.items() ]
                self._logger.debug(f'Calling paramiko.client.SSHClient.connect("{self._host}", {", ".join(paramiko_options_print)})')

                self._ssh_client = paramiko.client.SSHClient()
                self._ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self._ssh_client.connect(self._host, **paramiko_options)

                class ParamikoSocket:
                    def __init__(self, client, stdin, stdout, stderr):
                        self._client = client
                        self._stdin = stdin
                        self._stdout = stdout
                        self._stderr = stderr

                    def recv(self, size, flags):
                        return self._stdout.read(size)
                        
                    def recv_into(self, buffer, size, flags):
                        buffer[ : size] = self.recv(size, flags)
                        return size

                    def sendall(self, buffer):
                        self._stdin.write(buffer)
                        self._stdin.flush()

                    def close(self):
                        self._stdin.close()
                        self._stdout.close()
                        self._stderr.close()
                        self._client.close()

                self._logger.debug(f'Calling paramiko.client.SSHClient.exec_command("{command}")')
                stdin, stdout, stderr = self._ssh_client.exec_command(command)

                # If stderr is blocking, then we can't check it without locking the whole program up
                try:
                    stderr.channel.settimeout(0.5) # Give the subprocess time to connect

                    for line in stderr.readlines():
                        self._logger.warning(line.rstrip())

                except:
                    self._logger.debug('Unable to change SSH subprocess stderr to non-blocking')

                stdin.channel.settimeout(self._timeout)
                stdout.channel.settimeout(self._timeout)

                self._socket = ParamikoSocket(self._ssh_client, stdin, stdout, stderr)

        msg_login = Message(String(self._username))
        msg_login.ndims = 1
        msg_login.dims[0] = MDSIP_VERSION

        self._logger.debug(f'Sending login request with username="{self._username}"')
        self._send(msg_login)

        # The login response packet is a copy of the login request packet with a few fields changed,
        # so if we process it with self._recv it will wait for data that will never come
        msg_login_buffer = bytearray(ctypes.sizeof(msg_login))
        msg_login_view = memoryview(msg_login_buffer)
        while len(msg_login_view) > 0:
            bytes_read = self._socket.recv_into(msg_login_view, len(msg_login_view), 0)
            msg_login_view = msg_login_view[bytes_read : ]
        
        msg_login = Message.from_buffer_copy(msg_login_buffer)

        if STATUS_NOT_OK(msg_login.status):
            raise MdsException('Failed to login')

        self._compression_level = (msg_login.status & 0x1E) >> 1
        self._client_type = msg_login.client_type

        if msg_login.ndims > 0:
            self._server_version = msg_login.dims[0]

        self._logger.debug(f'Received login response with version={self._server_version} client_type={self._client_type} compression_level={self._compression_level}')

    def disconnect(self):
        """Close the socket."""

        if self._socket:
            self._logger.debug('Disconnecting')
            self._socket.close()
            self._socket = None

    def reconnect(self):
        """Call `disconnect()` and then `connect()`."""
        self.disconnect()
        self.connect()

    def _send(self, msg: Message):

        buffer = msg.pack()

        self._logger.debug(f'Sending packet with msglen={msg.msglen} dtype_id={dtype_to_string(msg.dtype_id)} length={msg.length} dimct={msg.ndims} dims={list(msg.dims)}')
        self._socket.sendall(buffer)

    def _recv(self):

        msg = Message()
        data = Descriptor()

        msg_buffer = bytearray(ctypes.sizeof(msg))
        msg_view = memoryview(msg_buffer)
        while len(msg_view) > 0:
            bytes_read = self._socket.recv_into(msg_view, len(msg_view), 0)
            msg_view = msg_view[bytes_read : ]

        msg = Message.from_buffer_copy(msg_buffer)

        self._logger.debug(f'Received message with msglen={msg.msglen} dtype_id={dtype_to_string(msg.dtype_id)} length={msg.length} dimct={msg.ndims} dims={list(msg.dims)}')

        data_length = msg.msglen - ctypes.sizeof(msg)
        if data_length > 0:
            data_buffer = bytearray(data_length)
            data_view = memoryview(data_buffer)
            while len(data_view) > 0:
                bytes_read = self._socket.recv_into(data_view, len(data_view), 0)
                data_view = data_view[bytes_read : ]

                self._logger.debug(f'Received data packet of {bytes_read} bytes, {data_length - len(data_view)}/{data_length}')

            data = msg.unpack_data(data_buffer)

        return msg, data

    def get(self, expr, *args):
        """
        Evaluate an expression on the remote server and return the result. This works like
        `mdsvalue()` in our other APIs.

        :param str expr: The TDI expression to be evaluated, possibly with `$` placeholders
        :param *args: The optional arguments to be inserted for the placeholders in the
            expression. All native python/numpy types will be converted to Descriptors.
        :return: The result of executing the expression.
        :rtype: :class:`Descriptor`
        :raises TimeoutError: if the connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the result status indicates an error.
        """

        if expr.strip() == '':
            return Descriptor()

        mget = Message(expr, compression_level=self._compression_level)
        mget.nargs = 1 + len(args)

        self._message_id += 1
        mget.message_id = self._message_id

        self._send(mget)

        for i, arg in enumerate(args):
            marg = Message(arg, compression_level=self._compression_level)
            marg.nargs = mget.nargs
            marg.message_id = mget.message_id
            marg.descriptor_idx = i + 1

            self._send(marg)

        manswer, data = self._recv()

        if STATUS_NOT_OK(manswer.status):
            raise getException(manswer.status)

        return data

    def getObject(self, expr, *args):
        """
        Evaluate a `get()` expression, but the expression will be wrapped in 'SerializeOut'
        and `deserialize()` will be called on the result. This allows you to retrieve data
        with CLASS_R or CLASS_APD, that otherwise cannot be retrieved with a regular `get()`.

        :param str expr: The TDI expression to be evaluated, possibly with `$` placeholders
        :param *args: The optional arguments to be inserted for the placeholders in the
            expression. All native python/numpy types will be converted to Descriptors.
        :return: The result of executing the expression.
        :rtype: :class:`Descriptor`
        :raises TimeoutError: if the connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the result status indicates an error.
        """
        return self.get(f'SerializeOut(`({expr};))', *args).deserialize(conn=self)

    def put(self, path, expr, *args):
        """
        Put an evaluated expression into a node in the last opened MDSplus tree.

        :param str path: The path to the node to write data into.
        :param str expr: The TDI expression to be evaluated, possibly with `$` placeholders
        :param *args: The optional arguments to be inserted for the placeholders in the
            expression. All native python/numpy types will be converted to Descriptors.
        :raises TimeoutError: if the connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the result status indicates an error.
        """
        args = [path, expr] + list(args)
        args_format = ','.join('$' * len(args))
        status = self.get(f'TreePut({args_format})', *args).data()

        if STATUS_NOT_OK(status):
            raise getException(status)

    def getMany(self):
        """
        Return a :class:`GetMany` object tied to this connection.

        :return: :class:`GetMany(self)`
        :rtype: :class:`GetMany`
        """

        return GetMany(self)

    def putMany(self):
        """
        Return a :class:`PutMany` object tied to this connection.

        :return: :class:`PutMany(self)`
        :rtype: :class:`PutMany`
        """

        if self.getServerVersion() < (7, 145, 7):
            import warnings
            warnings.warn('putMany is likely broken in MDSplus < 7.145.7, use with caution')

        return PutMany(self)

    def openTree(self, tree: str, shot: int):
        """
        Open an MDSplus tree on a remote server.

        :param str tree: The tree name to open.
        :param int shot: The shot number to open.
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the tree could not be opened.
        """

        status = self.get('TreeOpen($,$)', tree, shot).data()

        if STATUS_NOT_OK(status):
            raise getException(status)

    def closeTree(self, tree: str, shot: int):
        """
        Close an MDSplus tree on the remote server.

        :param str tree: The tree name to close.
        :param int shot: The shot number to close.
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the tree could not be closed.
        """

        status = self.get('TreeClose($,$)', tree, shot).data()

        if STATUS_NOT_OK(status):
            raise getException(status)

    def closeAllTrees(self):
        """
        Close all open MDSplus trees.

        :return: The number of trees closed.
        :rtype: Descriptor
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if there was a problem executing the `get()`.
        """

        return self.get("_i=0;WHILE(IAND(TreeClose(),1)) _i++;_i")

    def setDefault(self, path: str):
        """
        Change the current default tree location on the remote server

        :param str path: The tree node path to be set as the new default location.
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the tree node could not be found, or the location
            could not be changed.
        """

        status = self.get('TreeSetDefault($)', path).data()

        if STATUS_NOT_OK(status):
            raise getException(status)

    def tcl(self, command: str):
        """
        Execute a mdstcl command and return the result.

        :param str command: The mdstcl command to run.
        :return: The command output from mdstcl.
        :rtype: str
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if there was a problem executing the command.
        """
        result = self.get('Tcl($,_res);_res', command)
        if result is None:
            return ''
        return result.data()
    
    def getServerVersion(self):
        import re

        show_version = self.tcl('show version')
        matches = re.search(r'MDSplus version: ([0-9]+)\.([0-9]+)\.([0-9]+)', show_version, re.MULTILINE)

        return (int(matches[1]), int(matches[2]), int(matches[3]))

    def mdstcl(self):
        """
        Create a faux mdstcl prompt, calling `tcl()` for each command.

        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        """

        while True:
            # Use sys.stdout to avoid cluttering the python history with TCL commands
            sys.stdout.write('TCL> ')
            sys.stdout.flush()
            command = sys.stdin.readline()

            # Ctrl+D
            if len(command) == 0:
                print()
                break

            command = command.strip()

            # Empty command
            if len(command) == 0:
                continue

            if command == 'exit':
                break

            try:
                result = self.tcl(command)
            except MdsException as e:
                print(e)

            if result is not None:
                print(result, end='')

    def tdic(self):
        """
        Create a faux tdic prompt, calling `get()` for each command.

        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        """

        while True:
            # Use sys.stdout to avoid cluttering the python history with TCL commands
            sys.stdout.write('TDI> ')
            sys.stdout.flush()
            command = sys.stdin.readline()

            # Ctrl+D
            if len(command) == 0:
                print()
                break

            command = command.strip()

            if command == 'exit':
                break

            result = None
            try:
                result = self.get(command)
            except MdsException as e:
                print(e)

            if result is None:
                print(repr(dMISSING()))
            else:
                print(repr(result))

class GetMany:
    """
    Allows you to build a list of expressions to evaluate, reducing the number of network
    transactions needed.

    This must be constructed with a reference to a :class:`Connection`, this be done with
    `connection.getMany()`. You can then call `append()` to add expressions to the list.
    Once the list is complete, you must use `execute()` to send the expressions to the
    server and retrieve the result. This will return a dictionary of the result, or you
    can use `get()` to access the expressions by name.

    Example:
    ```
    gm = c.getMany()
    gm.append('y', '\\IP')
    gm.append('x', 'dim_of(\\IP)')
    result = gm.execute()

    y = gm.get('y')
    x = gm.get('x')
    # or
    y = result['y']
    x = result['x']
    ```
    """

    def __init__(self, connection: Connection):
        self._connection = connection
        self._queries = List()
        self._result = None

    def append(self, name, exp, *args):
        """
        Add a named expression to the list to be evaluated by `execute()`.

        :param str name: The name of the expression to evaluate. This will be how you
            retrieve the data from either the result dictionary or using `get()`.
        :param str expr: The TDI expression to be evaluated, possibly with `$` placeholders.
        :param *args: The optional arguments to be inserted for the placeholders in the
            expression. All native python/numpy types will be converted to Descriptors.
        """
        self._queries.append(Dictionary({
            'name': name,
            'exp': exp,
            'args': list(args),
        }))

    def remove(self, name):
        """
        Remove a named expression from the list

        :param str name: The name of the expression to remove
        """
        for query in self._queries:
            if query['name'] == name:
                self._queries.remove(query)
                break

    def execute(self):
        """
        Execute all expressions in the list by calling `GetManyExecute()` on the remote
        server, and passing the serialized list as data.

        :return: The Dictionary of results from the expressions. In the format of,
            `{ NAME: { 'value': DATA } }` if the expression succeeded, or
            `{ NAME: { 'error': ERROR_STRING } }` if there was an error.
        :rtype: :class:`Dictionary`
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the result of GetManyExecute() is an error string, or
            if `get()` encounters an error.
        """
        result = self._connection.get('GetManyExecute($)', self._queries.serialize())

        if isinstance(self._result, String):
            raise MdsException(f'GetMany Error: {self._result.data()}')

        self._result = result.deserialize()
        return self._result

    def get(self, name):
        """
        Get the result of a named expression, or raise an error if the evaluation failed.

        :param str name: The name of the expression
        :return: The resulting data from the expression
        :rtype: :class:`Descriptor`
        :raises MdsException: if `execute()` has not been called, or if the evaluation of
            the expression on the server failed.
        """
        if self._result is None:
            raise MdsException('GetMany has not been executed, call execute() first.')

        if name not in self._result:
            return None

        result = self._result[name]
        if 'value' in result:
            return result['value']

        raise getExceptionFromError(result['error'].data())

class PutMany:
    """
    Allows you to build a list of nodes with expressions to evaluate and store in them,
    reducing the number of network transactions needed.

    This must be constructed with a reference to a :class:`Connection`, this be done with
    `connection.putMany()`. You can then call `append()` to add nodes/expressions to the list.
    Once the list is complete, you must use `execute()` to send the expressions to the
    server and insert the data into each node. This will return a dictionary of the status of
    each node's operation, or you can use `checkStatus()` to check a given node.

    If you wish to store the result of an expression in a node, you must prefix the expression with '`'.

    Example:
    ```
    pm = c.putMany()
    pm.append('NODE_A', '$', 42)
    pm.append('NODE_B', '$', numpy.array([1, 2, 3, 4]))
    pm.append('NODE_C', 'SerializeIn($)', Signal(data, None, dim).serialize())
    result = pm.execute()

    gm.checkStatus('a')
    # or
    if result['a'] != 'Success':
        pass
    ```
    """


    def __init__(self, connection: Connection):
        self._connection = connection
        self._queries = List()
        self._result = Dictionary()

    def append(self, node, exp, *args):
        """
        Add a node/expression to the list to be evaluated and inserted by `execute()`

        :param str node: The node to insert the evaluated expression into.
        :param str expr: The TDI expression to be evaluated, possibly with `$` placeholders
        :param *args: The optional arguments to be inserted for the placeholders in the
            expression. All native python/numpy types will be converted to Descriptors
        """
        self._queries.append(Dictionary({
            'node': node,
            'exp': exp,
            'args': list(args),
        }))

    def remove(self, node):
        """
        Remove a node/expression from the list

        :param str name: The node to remove from the list
        """
        for query in self._queries:
            if query['node'] == node:
                self._queries.remove(query)
                break

    def execute(self):
        """
        Execute and insert all expressions in the list by calling `PutManyExecute()`
        on the remote server, and passing the serialized list as data.

        :return: The Dictionary of results from the expressions. In the format of,
            `{ NAME: 'Success' }` if the expression succeeded, or
            `{ NAME: ERROR_STRING } }` if there was an error.
        :rtype: :class:`Dictionary`
        :raises TimeoutError: if the network connection fails.
        :raises BrokenPipeError: if the SSH subprocess fails.
        :raises OSError: if the paramiko client fails.
        :raises MdsException: if the result of PutManyExecute() is an error string, or
            if `get()` encounters an error.
        """
        result = self._connection.get('PutManyExecute($)', self._queries.serialize())

        if isinstance(self._result, String):
            raise MDSplusException(f'PutMany Error: {self._result.data()}')

        self._result = result.deserialize(conn=self)
        return self._result

    def checkStatus(self, node):
        """
        Check the status of inserting the data into a given node.

        :param str name: The node to check
        :return: The resulting "Success" string for the node
        :rtype: :class:`Descriptor`
        :raises MdsException: if `execute()` has not been called, or if the evaluation or
            insertion of the expression on the server failed.
        """
        if self._result is None:
            raise MdsException('PutMany has not been executed, call execute() first.')

        if node not in self._result:
            return None

        result = self._result[node]
        if self.result[node] != "Success":
            raise getExceptionFromError(self._result[node])

        return self.result[node]


#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import ctypes
import numpy

from .dtypedef import *

###
### VMS
###

class F_Floating(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('fraction', ctypes.c_uint32, 7),
        ('exponent', ctypes.c_uint32, 8),
        ('sign', ctypes.c_uint32, 1),
        ('fraction2', ctypes.c_uint32, 16),
    ]

class D_Floating(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('fraction', ctypes.c_uint64, 7),
        ('exponent', ctypes.c_uint64, 8),
        ('sign', ctypes.c_uint64, 1),
        ('fraction2', ctypes.c_uint64, 16),
        ('fraction3', ctypes.c_uint64, 16),
        ('fraction4', ctypes.c_uint64, 16),
    ]

class G_Floating(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('fraction', ctypes.c_uint64, 4),
        ('exponent', ctypes.c_uint64, 11),
        ('sign', ctypes.c_uint64, 1),
        ('fraction2', ctypes.c_uint64, 16),
        ('fraction3', ctypes.c_uint64, 16),
        ('fraction4', ctypes.c_uint64, 16),
    ]

class H_Floating(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('exponent', ctypes.c_uint64, 15),
        ('sign', ctypes.c_uint64, 1),
        ('fraction', ctypes.c_uint64, 16),
        ('fraction2', ctypes.c_uint64, 16),
        ('fraction3', ctypes.c_uint64, 16),
        ('fraction4', ctypes.c_uint64, 16),
        ('fraction5', ctypes.c_uint64, 16),
        ('fraction6', ctypes.c_uint64, 16),
        ('fraction7', ctypes.c_uint64, 16),
    ]

###
### IEEE
###

class IEEE_Single(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('mantissa', ctypes.c_uint32, 23),
        ('exponent', ctypes.c_uint32, 8),
        ('sign', ctypes.c_uint32, 1),
    ]

class IEEE_Double(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('mantissa', ctypes.c_uint64, 52),
        ('exponent', ctypes.c_uint64, 11),
        ('sign', ctypes.c_uint64, 1),
    ]

class IEEE_LongDouble(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('mantissa', ctypes.c_uint64, 64),
        ('mantissa2', ctypes.c_uint64, 48),
        ('exponent', ctypes.c_uint64, 15),
        ('sign', ctypes.c_uint64, 1),
    ]

###
### Unions
###

class Float(ctypes.Union):
    _pack_ = 1
    _fields_ = [
        ('F', F_Floating),
        ('FS', IEEE_Single),
        ('value', ctypes.c_float),
    ]

    def convert_F(self):
        
        sign = self.F.sign
        exponent = self.F.exponent - 128
        fraction = (
            (self.F.fraction << 16) |
            self.F.fraction2
        )

        self.FS.sign = sign
        self.FS.exponent = exponent + 126
        self.FS.mantissa = fraction

        return self.value

class Double(ctypes.Union):
    _pack_ = 1
    _fields_ = [
        ('D', D_Floating),
        ('G', G_Floating),
        ('FT', IEEE_Double),
        ('value', ctypes.c_double),
    ]

    def convert_D(self):
        
        sign = self.D.sign
        fraction = ((self.D.fraction << 48)
                    | (self.D.fraction2 << 32)
                    | (self.D.fraction3 << 16)
                    | self.D.fraction4)
        exponent = self.D.exponent - 128

        # TODO: What the fuck

        self.FT.sign = sign
        self.FT.exponent = exponent + 1022
        self.FT.mantissa = fraction

        return self.value

    def convert_G(self):

        sign = self.G.sign
        fraction = ((self.G.fraction << 48)
                    | (self.G.fraction2 << 32)
                    | (self.G.fraction3 << 16)
                    | self.G.fraction4)
        exponent = self.G.exponent - 1024

        self.FT.sign = sign
        self.FT.exponent = exponent + 1022
        self.FT.mantissa = fraction

        return self.value
    
class LongDouble(ctypes.Union):
    _pack_ = 1
    _fields_ = [
        ('H', H_Floating),
        ('FQ', IEEE_LongDouble),
        ('value', ctypes.c_longdouble),
    ]

    def convert_H(self):

        sign = self.H.sign
        # fraction = (self.H.fraction << 48) | (self.H.fraction2 << 32) | (self.H.fraction3 << 16) | self.H.fraction4
        # fraction2 = (self.H.fraction5 << 32) | (self.H.fraction6 << 16) | self.H.fraction7
        fraction = ((self.H.fraction << 96)
                    | (self.H.fraction2 << 80)
                    | (self.H.fraction3 << 64)
                    | (self.H.fraction4 << 48)
                    | (self.H.fraction5 << 32)
                    | (self.H.fraction6 << 16)
                    | self.H.fraction7)
        exponent = self.H.exponent - 1024

        self.FQ.sign = sign
        self.FQ.exponent = exponent + 1022
        self.FQ.mantissa = (fraction >> 48)
        self.FQ.mantissa2 = (fraction & 0xFFFFFF)

        return self.value

def convert_float(dtype, buffer):

    if dtype == DTYPE_F:
        return numpy.float32(Float.from_buffer(buffer).convert_F())

    elif dtype == DTYPE_D:
        return numpy.float64(Double.from_buffer(buffer).convert_D())

    elif dtype == DTYPE_G:
        return numpy.float64(Double.from_buffer(buffer).convert_G())

    elif dtype == DTYPE_H:
        # not supported
        # return numpy.float128(LongDouble.from_buffer(buffer).convert_H())
        return None

def convert_float_array(dtype, buffer):

    if dtype == DTYPE_F:
        length = len(buffer) // ctypes.sizeof(Float)
        items = (Float * length).from_buffer(buffer)
        return numpy.array([ item.convert_F() for item in items ], dtype=numpy.float32)
    
    elif dtype == DTYPE_D:
        length = len(buffer) // ctypes.sizeof(Double)
        items = (Double * length).from_buffer(buffer)
        return numpy.array([ item.convert_D() for item in items ], dtype=numpy.float64)
    
    elif dtype == DTYPE_G:
        length = len(buffer) // ctypes.sizeof(Double)
        items = (Double * length).from_buffer(buffer)
        return numpy.array([ item.convert_G() for item in items ], dtype=numpy.float64)


#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import numpy
import unittest

from ..descriptors import *

class DescriptorsTest(unittest.TestCase):

    def test_integer_limits(self):

        tests = [
            { 'type': UInt8,  'min': 0, 'max': numpy.iinfo(numpy.uint8).max },
            { 'type': UInt16, 'min': 0, 'max': numpy.iinfo(numpy.uint16).max },
            { 'type': UInt32, 'min': 0, 'max': numpy.iinfo(numpy.uint32).max },
            { 'type': UInt64, 'min': 0, 'max': numpy.iinfo(numpy.uint64).max },

            { 'type': Int8,  'min': numpy.iinfo(numpy.int8).min,  'max': numpy.iinfo(numpy.int8).max },
            { 'type': Int16, 'min': numpy.iinfo(numpy.int16).min, 'max': numpy.iinfo(numpy.int16).max },
            { 'type': Int32, 'min': numpy.iinfo(numpy.int32).min, 'max': numpy.iinfo(numpy.int32).max },
            { 'type': Int64, 'min': numpy.iinfo(numpy.int64).min, 'max': numpy.iinfo(numpy.int64).max },
        ]

        for info in tests:
            name = repr(info['type'])
            with self.subTest(name):

                data = info['type'](info['min'])
                self.assertEqual(type(data), info['type'])
                self.assertEqual(data.data(), info['min'])

                data = info['type'](info['min'])
                self.assertEqual(type(data), info['type'])
                self.assertEqual(data.data(), info['min'])

                data = info['type'](42)
                self.assertEqual(type(data), info['type'])
                self.assertEqual(data.data(), 42)

    def test_string(self):

        string = 'Hello, World!'
        data = String(string)
        self.assertEqual(type(data), String)
        self.assertEqual(data.data(), string)

        string = 'a' * 1048576 # 1MB
        data = String(string)
        self.assertEqual(type(data), String)
        self.assertEqual(data.data(), string)

        data = String()
        self.assertEqual(type(data), String)
        self.assertEqual(data.data(), '')

        self.assertRaises(UnicodeEncodeError, String, '🔥')

        buffer = '🔥'.encode()
        self.assertRaises(UnicodeDecodeError, String.unpack_data, buffer)

    def test_string_array(self):
        
        strings = ['a', 'b', 'c']
        data = StringArray(strings)
        self.assertEqual(type(data), StringArray)
        self.assertEqual(data, numpy.array(strings, dtype=str))
        self.assertTrue((data.data_raw() == ['a', 'b', 'c']).all())
        
        strings = ['one', 'seven', 'nineteen']
        data = StringArray(strings)
        self.assertEqual(type(data), StringArray)
        self.assertEqual(data, numpy.array(strings, dtype=str))
        self.assertTrue((data.data_raw() == ['one     ', 'seven   ', 'nineteen']).all())
        
        strings = [ ['one', 'two'], ['three', 'four'] ]
        data = StringArray(strings)
        self.assertEqual(type(data), StringArray)
        self.assertEqual(data, numpy.array(strings, dtype=str))
        self.assertTrue((data.data_raw() == [ ['one  ', 'two  '], ['three', 'four '] ]).all())

    def test_native_types(self):

        data = Descriptor('test')
        self.assertEqual(type(data), String)
        self.assertEqual(data, 'test')
        
        data = Descriptor(True)
        self.assertEqual(type(data), UInt8)
        self.assertEqual(data, True)

        data = Descriptor(42)
        self.assertEqual(type(data), Int64)
        self.assertEqual(data, 42)
        
        data = Descriptor(3.14159)
        self.assertEqual(type(data), Float64)
        self.assertEqual(data, 3.14159)
        
        data = Descriptor(b'\xCA\xFE')
        self.assertEqual(type(data), UInt8Array)
        self.assertEqual(data, numpy.array([ 0xCA, 0xFE ], dtype=numpy.uint8))

        data = Descriptor(bytearray(b'\xCA\xFE'))
        self.assertEqual(type(data), UInt8Array)
        self.assertEqual(data, numpy.array([ 0xCA, 0xFE ], dtype=numpy.uint8))

        buffer = b'\xCA\xFE'
        data = Descriptor(memoryview(buffer))
        self.assertEqual(type(data), UInt8Array)
        self.assertEqual(data, numpy.array([ 0xCA, 0xFE ], dtype=numpy.uint8))

        items = [ 1, 'b', 3.0 ]
        data = Descriptor(items)
        self.assertEqual(type(data), List)
        self.assertEqual(type(data[0]), Int64)
        self.assertEqual(type(data[1]), String)
        self.assertEqual(type(data[2]), Float64)
        self.assertListEqual(data.data(), items)

        items = ( 1, 'b', 3.0 )
        data = Descriptor(items)
        self.assertEqual(type(data), Tuple)
        self.assertEqual(type(data[0]), Int64)
        self.assertEqual(type(data[1]), String)
        self.assertEqual(type(data[2]), Float64)
        self.assertTupleEqual(data.data(), items)

    def test_vms_floats(self):
        
        tests = [
            {
                'dtype': DTYPE_F,
                'buffer': bytearray([ 0x49, 0x41, 0xD0, 0x0F ]),
                'type': Float32,
                'value': 3.14159,
            },
            # {
            #     'dtype': DTYPE_FC,
            #     'buffer': bytes([ 0x23, 0xC6, 0xCD, 0x70 ]),
            #     'type': Float32,
            #     'value': 3.14159,
            # },
            # {
            #     'dtype': DTYPE_D,
            #     'buffer': bytes([ 0x49, 0x41, 0xCF, 0x0F, 0xDC, 0x80, 0x70, 0x33 ]),
            #     'type': Float64,
            #     'value': 3.14159,
            # },
            # {
            #     'dtype': DTYPE_DC,
            #     'buffer': bytes([ 0x49, 0x41, 0xCF, 0x0F, 0xDC, 0x80, 0x70, 0x33 ]),
            #     'type': Float64,
            #     'value': 3.14159,
            # },
            {
                'dtype': DTYPE_G,
                'buffer': bytes([ 0x29, 0x40, 0xF9, 0x21, 0x1B, 0xF0, 0x6E, 0x86 ]),
                'type': Float64,
                'value': 3.14159,
            },
            # {
            #     'dtype': DTYPE_H,
            #     'buffer': bytes([ 0x02, 0x40, 0x1F, 0x92, 0x01, 0x9F, 0x66, 0xB8, 0x00, 0xE0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ]),
            #     'type': Float128,
            #     'value': 3.14159,
            # },
        ]


        for info in tests:
            name = dtype_to_string(info['dtype'])
            with self.subTest(name):

                dsc = mdsdsc_s_t(
                    length=len(info['buffer']),
                    dtype_id=info['dtype'],
                    class_id=CLASS_S,
                )

                buffer = bytes(dsc) + info['buffer']
                data = Descriptor.unpack(buffer)
                self.assertEqual(type(data), info['type'])
                self.assertAlmostEqual(data.data(), info['value'], places=5)
        
    def test_numpy_types(self):

        INTEGER_ARRAY = [ [2, 4], [6, 8], [16, 32] ]
        FLOAT_ARRAY = [ [1.0, 1.5], [2.0, 2.5], [3.0, 3.5] ]

        tests = [
            { 'type': UInt8,        'value': numpy.uint8(42), },
            { 'type': UInt16,       'value': numpy.uint16(42), },
            { 'type': UInt32,       'value': numpy.uint32(42), },
            { 'type': UInt64,       'value': numpy.uint64(42), },
            { 'type': Int8,         'value': numpy.int8(42), },
            { 'type': Int16,        'value': numpy.int16(42), },
            { 'type': Int32,        'value': numpy.int32(42), },
            { 'type': Int64,        'value': numpy.int64(42), },
            { 'type': Float32,      'value': numpy.float32(3.14159), },
            { 'type': Float64,      'value': numpy.float64(3.14159), },

            { 'type': UInt8Array,   'value': numpy.array(INTEGER_ARRAY, dtype=numpy.uint8) },
            { 'type': UInt16Array,  'value': numpy.array(INTEGER_ARRAY, dtype=numpy.uint16) },
            { 'type': UInt32Array,  'value': numpy.array(INTEGER_ARRAY, dtype=numpy.uint32) },
            { 'type': UInt64Array,  'value': numpy.array(INTEGER_ARRAY, dtype=numpy.uint64) },
            { 'type': Int8Array,    'value': numpy.array(INTEGER_ARRAY, dtype=numpy.int8) },
            { 'type': Int16Array,   'value': numpy.array(INTEGER_ARRAY, dtype=numpy.int16) },
            { 'type': Int32Array,   'value': numpy.array(INTEGER_ARRAY, dtype=numpy.int32) },
            { 'type': Int64Array,   'value': numpy.array(INTEGER_ARRAY, dtype=numpy.int64) },
            { 'type': Float32Array, 'value': numpy.array(FLOAT_ARRAY,   dtype=numpy.float32) },
            { 'type': Float64Array, 'value': numpy.array(FLOAT_ARRAY,   dtype=numpy.float64) },

            # TODO: string
            # TODO: string array
        ]

        for info in tests:
            name = repr(info['value'].dtype)
            with self.subTest(name):

                data = Descriptor(info['value'])
                self.assertEqual(type(data), info['type'])
                self.assertEqual(data.data().dtype, info['value'].dtype)
                self.assertEqual(data, info['value'])

    def test_apd_types(self):

        # List

        data = List()
        self.assertEqual(data, list())
        self.assertEqual(data.data(), list())
        
        data = List(1, 2, 3)
        self.assertEqual(data, [1, 2, 3])
        self.assertEqual(data.data(), [1, 2, 3])
        
        data.append(4)
        self.assertEqual(data, [1, 2, 3, 4])
        self.assertEqual(data.data(), [1, 2, 3, 4])

        data[1] = 5
        self.assertEqual(data, [1, 5, 3, 4])
        self.assertEqual(data.data(), [1, 5, 3, 4])
        
        data = List([1, 2], [3, 4], [5, 6])
        self.assertEqual(data, [ [1, 2], [3, 4], [5, 6] ])
        self.assertEqual(data.data(), [ [1, 2], [3, 4], [5, 6] ])

        # Tuple

        data = Tuple()
        self.assertEqual(data, tuple())
        self.assertEqual(data.data(), tuple())
        
        data = Tuple(1, 2, 3)
        self.assertEqual(data, (1, 2, 3))
        self.assertEqual(data.data(), (1, 2, 3))

        # Dictionary

        data = Dictionary()
        self.assertEqual(data, dict())
        self.assertEqual(data.data(), dict())

        data = Dictionary({ 'a': 1, 'b': 2, 'c': 3 })
        self.assertEqual(data, { 'a': 1, 'b': 2, 'c': 3 })
        self.assertEqual(data.data(), { 'a': 1, 'b': 2, 'c': 3 })

        data = Dictionary(('a', 1), ('b', 2), ('c', 3))
        self.assertEqual(data, { 'a': 1, 'b': 2, 'c': 3 })
        self.assertEqual(data.data(), { 'a': 1, 'b': 2, 'c': 3 })
        
        data['d'] = 4
        self.assertEqual(data, { 'a': 1, 'b': 2, 'c': 3, 'd': 4 })
        self.assertEqual(data.data(), { 'a': 1, 'b': 2, 'c': 3, 'd': 4 })

        data['b'] = 5
        self.assertEqual(data, { 'a': 1, 'b': 5, 'c': 3, 'd': 4 })
        self.assertEqual(data.data(), { 'a': 1, 'b': 5, 'c': 3, 'd': 4 })

    def test_arrays(self):
        
        INTEGER_ARRAY = [ 2, 4, 6 ]
        FLOAT_ARRAY = [ 1.0, 1.1, 1.2 ]

        tests = [
            { 'value': UInt32Array(list(range(100))), 'length': 100 },

            { 'value': UInt8Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': UInt16Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': UInt32Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': UInt64Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': Int8Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': Int16Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': Int32Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': Int64Array(INTEGER_ARRAY), 'length': 3 },
            { 'value': Float32Array(FLOAT_ARRAY), 'length': 3 },
            { 'value': Float64Array(FLOAT_ARRAY), 'length': 3 },

            { 'value': StringArray(['one', 'seven', 'nineteen']), 'length': 3 },
        ]

        for info in tests:
            name = repr(type(info['value'])) + f"({info['length']})"
            with self.subTest(name):

                self.assertEqual(len(info['value']), info['length'])

                data = info['value'].data()
                
                for i in range(info['length']):
                    self.assertEqual(info['value'][i], data[i])
                
                for a, b in zip(info['value'], data):
                    self.assertEqual(a, b)

    def test_record_types(self):

        # Signal

        data = Signal()
        self.assertEqual(data.value, None)
        self.assertEqual(data.raw, None)
        self.assertEqual(data.dimensions, [])
        self.assertEqual(data.data(), None)
        self.assertRaises(IndexError, data.dim_of)

        data = Signal('value', 'raw', 'dim0', 'dim1', 'dim2')
        self.assertEqual(data.value, 'value')
        self.assertEqual(data.raw, 'raw')
        self.assertListEqual(list(data.dimensions), ['dim0', 'dim1', 'dim2'])
        self.assertEqual(data.data(), 'value')
        self.assertEqual(data.dim_of(), 'dim0')
        self.assertEqual(data.dim_of(0), 'dim0')
        self.assertEqual(data.dim_of(1), 'dim1')
        self.assertEqual(data.dim_of(2), 'dim2')
        self.assertEqual(data.dim_of(-1), 'dim2')
        self.assertRaises(IndexError, data.dim_of, 3)

        data = Signal(None, 'raw', 'dim')
        self.assertEqual(data.value, None)
        self.assertEqual(data.raw, 'raw')
        self.assertListEqual(list(data.dimensions), ['dim'])
        self.assertEqual(data.data(), 'raw')
        self.assertEqual(data.dim_of(), 'dim')
        self.assertRaises(IndexError, data.dim_of, 1)

        # Dimension

        data = Dimension()
        self.assertEqual(data.window, None)
        self.assertEqual(data.axis, None)

        data = Dimension('window', 'axis')
        self.assertEqual(data.window, 'window')
        self.assertEqual(data.axis, 'axis')

        # TODO: Put in a realistic axis
        data = Dimension(Window(0, 10, 5), 'axis')
        self.assertEqual(data.window, Window(0, 10, 5))
        self.assertEqual(data.axis, 'axis')

        # Window

        data = Window()
        self.assertEqual(data.startidx, None)
        self.assertEqual(data.endingidx, None)
        self.assertEqual(data.value_at_idx0, None)

        data = Window('startidx', 'endingidx', 'value_at_idx0')
        self.assertEqual(data.startidx, 'startidx')
        self.assertEqual(data.endingidx, 'endingidx')
        self.assertEqual(data.value_at_idx0, 'value_at_idx0')

        # Slope

        data = Slope()
        self.assertEqual(data.segments, [])

        self.assertRaises(Exception, Slope, 'broken')
        self.assertRaises(Exception, Slope, 'broken1', 'broken2')

        data = Slope('slope', 'begin', 'ending')
        self.assertListEqual(list(data.segments), [('slope', 'begin', 'ending')])

        data = Slope('slope0', 'begin0', 'ending0', 'slope1', 'begin1', 'ending1')
        self.assertListEqual(list(data.segments), [('slope0', 'begin0', 'ending0'), ('slope1', 'begin1', 'ending1')])

        # Function

        data = Function()
        self.assertEqual(data.opcode, 0)
        self.assertEqual(data.arguments, [])

        data = Function(162, None, 'whoami') # TdiCompile("whoami()")
        self.assertEqual(data.opcode, 162)
        self.assertEqual(data.arguments, [None, 'whoami'])

        self.assertRaises(Exception, data.data)

        # Conglom

        data = Conglom()
        self.assertEqual(data.image, None)
        self.assertEqual(data.model, None)
        self.assertEqual(data.name, None)
        self.assertEqual(data.qualifiers, None)

        data = Conglom('image', 'model', 'name', 'qualifiers')
        self.assertEqual(data.image, 'image')
        self.assertEqual(data.model, 'model')
        self.assertEqual(data.name, 'name')
        self.assertEqual(data.qualifiers, 'qualifiers')

        # Range

        data = Range()
        self.assertEqual(data.begin, None)
        self.assertEqual(data.ending, None)
        self.assertEqual(data.deltaval, None)

        data = Range('begin', 'ending', 'deltaval')
        self.assertEqual(data.begin, 'begin')
        self.assertEqual(data.ending, 'ending')
        self.assertEqual(data.deltaval, 'deltaval')

        data = Range(0, 10, 1)
        self.assertEqual(data.data(), range(0, 10, 1))

        # Python range() only supports integers
        data = Range(0.0, 1.0, 0.1)
        self.assertRaises(TypeError, data.data)

        # Action

        data = Action()
        self.assertEqual(data.dispatch, None)
        self.assertEqual(data.task, None)
        self.assertEqual(data.errorlogs, None)
        self.assertEqual(data.completion_message, None)
        self.assertEqual(data.performance, None)

        data = Action('dispatch', 'task', 'errorlogs', 'completion_message', 'performance')
        self.assertEqual(data.dispatch, 'dispatch')
        self.assertEqual(data.task, 'task')
        self.assertEqual(data.errorlogs, 'errorlogs')
        self.assertEqual(data.completion_message, 'completion_message')
        self.assertEqual(data.performance, 'performance')

        data = Action(Dispatch(0, 'ANALYSIS', 'INIT', 999, ''), Function(162, None, 'DoAnalysis'), '', None, None)
        self.assertEqual(data.dispatch, Dispatch(0, 'ANALYSIS', 'INIT', 999, ''))
        self.assertEqual(data.task, Function(162, None, 'DoAnalysis'))
        self.assertEqual(data.errorlogs, '')
        self.assertEqual(data.completion_message, None)
        self.assertEqual(data.performance, None)

        # Dispatch

        data = Dispatch()
        self.assertEqual(data.treesched, 0)
        self.assertEqual(data.ident, None)
        self.assertEqual(data.phase, None)
        self.assertEqual(data.when, None)
        self.assertEqual(data.completion, None)

        # TreeSCHED_ASYNC = 1
        data = Dispatch(1, 'ident', 'phase', 'when', 'completion')
        self.assertEqual(data.treesched, 1)
        self.assertEqual(data.ident, 'ident')
        self.assertEqual(data.phase, 'phase')
        self.assertEqual(data.when, 'when')
        self.assertEqual(data.completion, 'completion')

        # Program

        data = Program()
        self.assertEqual(data.time_out, None)
        self.assertEqual(data.program, None)

        data = Program('time_out', 'program')
        self.assertEqual(data.time_out, 'time_out')
        self.assertEqual(data.program, 'program')

        # Routine

        data = Routine()
        self.assertEqual(data.time_out, None)
        self.assertEqual(data.image, None)
        self.assertEqual(data.routine, None)
        self.assertEqual(data.arguments, [])

        data = Routine('time_out', 'image', 'routine', 'arg0', 'arg1', 'arg2')
        self.assertEqual(data.time_out, 'time_out')
        self.assertEqual(data.image, 'image')
        self.assertEqual(data.routine, 'routine')
        self.assertListEqual(list(data.arguments), ['arg0', 'arg1', 'arg2'])

        # Procedure

        data = Procedure()
        self.assertEqual(data.time_out, None)
        self.assertEqual(data.language, None)
        self.assertEqual(data.procedure, None)
        self.assertEqual(data.arguments, [])

        data = Procedure('time_out', 'language', 'procedure', 'arg0', 'arg1', 'arg2')
        self.assertEqual(data.time_out, 'time_out')
        self.assertEqual(data.language, 'language')
        self.assertEqual(data.procedure, 'procedure')
        self.assertListEqual(list(data.arguments), ['arg0', 'arg1', 'arg2'])

        # Method

        data = Method()
        self.assertEqual(data.time_out, None)
        self.assertEqual(data.method, None)
        self.assertEqual(data.device, None)

        data = Method('time_out', 'method', 'device')
        self.assertEqual(data.time_out, 'time_out')
        self.assertEqual(data.method, 'method')
        self.assertEqual(data.device, 'device')

        # Dependency

        data = Dependency()
        self.assertEqual(data.treedep, 0)
        self.assertEqual(data.arguments, [])

        # TreeDEPENDENCY_AND = 10
        data = Dependency(10, 'first', 'second')
        self.assertEqual(data.treedep, 10)
        self.assertListEqual(list(data.arguments), ['first', 'second'])

        # Condition

        data = Condition()
        self.assertEqual(data.treecond, 0)
        self.assertEqual(data.condition, None)

        # TreeNEGATE_CONDITION = 7
        data = Condition(7, 'condition')
        self.assertEqual(data.treecond, 7)
        self.assertEqual(data.condition, 'condition')

        # WithUnits

        data = WithUnits()
        self.assertEqual(data.value, None)
        self.assertEqual(data.units, None)
        self.assertEqual(data.data(), None)

        data = WithUnits('value', 'units')
        self.assertEqual(data.value, 'value')
        self.assertEqual(data.units, 'units')
        self.assertEqual(data.data(), 'value')

        data = WithUnits(60.0, 'Hz')
        self.assertEqual(data.value, 60.0)
        self.assertEqual(data.units, 'Hz')
        self.assertEqual(data.data(), 60.0)

        data = WithUnits(Signal([1, 2, 3], None, [0.0, 0.1, 0.2]), 'counts')
        self.assertEqual(data.value, Signal([1, 2, 3], None, [0.0, 0.1, 0.2]))
        self.assertEqual(data.units, 'counts')
        self.assertEqual(data.data(), [1, 2, 3])

        # Call

        data = Call()
        self.assertEqual(data.return_dtype_id, DTYPE_L)
        self.assertEqual(data.image, None)
        self.assertEqual(data.routine, None)
        self.assertEqual(data.arguments, [])

        data = Call(DTYPE_MISSING, 'image', 'routine', 'arg0', 'arg1', 'arg2') # void
        self.assertEqual(data.return_dtype_id, DTYPE_MISSING)
        self.assertEqual(data.image, 'image')
        self.assertEqual(data.routine, 'routine')
        self.assertListEqual(list(data.arguments), ['arg0', 'arg1', 'arg2'])

        # WithError

        data = WithError()
        self.assertEqual(data.value, None)
        self.assertEqual(data.error, None)
        self.assertEqual(data.data(), None)
        self.assertEqual(data.getException(), None)

        data = WithError('value', 'error')
        self.assertEqual(data.value, 'value')
        self.assertEqual(data.error, 'error')
        self.assertEqual(data.data(), 'value')

        data = WithError(-1, '%TDI-E-INVCLADTY, Invalid mixture of storage class and data type')
        self.assertEqual(data.value, -1)
        self.assertEqual(data.error, '%TDI-E-INVCLADTY, Invalid mixture of storage class and data type')
        self.assertEqual(data.data(), -1)
        self.assertEqual(data.getException(), TdiINVCLADTY)

        # Opaque

        data = Opaque()
        self.assertEqual(data.value, None)
        self.assertEqual(data.opaque_type, None)
        self.assertEqual(data.data(), None)

        data = Opaque('value', 'opaque_type')
        self.assertEqual(data.value, 'value')
        self.assertEqual(data.opaque_type, 'opaque_type')
        self.assertEqual(data.data(), 'value')



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import unittest

from ..exceptions import *

class ExceptionsTest(unittest.TestCase):

    def test_status(self):

        self.assertTrue(STATUS_OK(TreeSUCCESS.status))
        self.assertFalse(STATUS_NOT_OK(TreeSUCCESS.status))
        
        self.assertTrue(STATUS_NOT_OK(DevCAMERA_NOT_FOUND.status))
        self.assertFalse(STATUS_OK(DevCAMERA_NOT_FOUND.status))

        self.assertEqual(STATUS_FACILITY(TreeNODATA.status), 4049)
        self.assertEqual(STATUS_MESSAGE(TreeNODATA.status), 4124)
        self.assertEqual(STATUS_SEVERITY(TreeNODATA.status), 2)

    def test_get_exception(self):

        self.assertEqual(getException(65545), MDSplusSUCCESS)

        self.assertEqual(getException(265388200), TreeNOT_OPEN)

        unknown_exception = getException(1234) 
        self.assertEqual(type(unknown_exception), MdsException)
        self.assertEqual(str(unknown_exception), 'Unknown status: 1234')

    def test_get_exception_from_error(self):
        
        self.assertEqual(getExceptionFromError("%TREE-W-NOT_OPEN, Tree not currently open"), TreeNOT_OPEN)
        
        self.assertEqual(getExceptionFromError("%TREE-W-NOT_OPEN, This part doesn't matter"), TreeNOT_OPEN)
        
        unknown_exception = getExceptionFromError("This part definitely matters")
        self.assertEqual(type(unknown_exception), MdsException)
        self.assertEqual(str(unknown_exception), 'This part definitely matters')
